/***************************************************************************
 *   Copyright (C) 2012 by Graziano Giuliani                               *
 *   ggiulian at ictp.it                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details. (see COPYING)            *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   LIC: GPL                                                              *
 *                                                                         *
 ***************************************************************************/

#include <string>
#include <cmath>
#include <libgen.h>
#include <netcdf>
#include <ctime>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

using namespace netCDF;
using namespace netCDF::exceptions;

#define NC_ERR 2

#define SIZEOFDATE 64

int main(int argc, char *argv[])
{

  if ( argc < 3 )
  {
    std::cerr << "Not enough arguments" << std::endl;
    std::cerr << "Usage" << std::endl;
    std::cerr << basename(argv[0]) << " nx nt outfile" << std::endl;
    std::cerr << "where: " << std::endl;
    std::cerr << "\tnx is the size of the problem" << std::endl;
    std::cerr << "\tnt is the number of timesteps" << std::endl;
    std::cerr << "\toutfile is the name of the output file" << std::endl;
    return 1;
  }

  int nx = (int) strtol(argv[1], NULL, 0);
  int nt = (int) strtol(argv[2], NULL, 0);
  if ( nx < 1 || nt < 1 )
    throw "NX must be greater than one";
  int hx = nx/2;

  std::cout << "NX = " << nx << std::endl;
  std::cout << "NT = " << nt << std::endl;

  NcFile newf(argv[3], NcFile::replace);

  time_t xt = time(NULL);
  struct tm *tmp = localtime(&xt);
  if (tmp == NULL)
    throw "Error in get system date";
  char datetime[SIZEOFDATE];
  (void) strftime(datetime, SIZEOFDATE, "Created on %c", tmp);

  try {
    newf.putAtt("title","Random walk of a log surface");
    newf.putAtt("history",datetime);
  }
  catch(NcException& e)
  {
    e.what(); 
    return NC_ERR;
  }

  NcDim xdim , ydim , tdim;
  try {
    xdim = newf.addDim("X", nx);
    ydim = newf.addDim("Y", nx);
    tdim = newf.addDim("time");
  }
  catch(NcException& e)
  {
    e.what(); 
    return NC_ERR;
  }

  NcVar xvar;
  try {
    std::vector<NcDim> dims;
    dims.push_back(tdim);
    dims.push_back(ydim);
    dims.push_back(xdim);
    xvar = newf.addVar("randomwalk", ncDouble, dims);
  }
  catch(NcException& e)
  {
    e.what(); 
    return NC_ERR;
  }

  double *matrix = new double[nx*nx];
  gsl_rng_env_setup();
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r = gsl_rng_alloc(T);

  double x = 0.0;
  double y = 0.0;
  double dx, dy;
  std::vector<size_t> startp , countp;
  startp.push_back((size_t) 0);
  startp.push_back((size_t) 0);
  startp.push_back((size_t) 0);
  countp.push_back((size_t) 1);
  countp.push_back((size_t) nx);
  countp.push_back((size_t) nx);
  for (size_t it = 0; it < (size_t) nt; it++)
  {
    gsl_ran_dir_2d(r, &dx, &dy);
    x += dx;
    y += dy;
    for (int i = 0; i < nx; i++)
    {
      for (int j = 0; j < nx; j++)
      {
        matrix[j*nx+i] = log(pow((double)(i-hx)-x,2.0) + 
                             pow((double)(j-hx)-y,2.0)+0.1);
      }
    }
    try {
      startp[0] = it;
      xvar.putVar(startp,countp,matrix);
    }
    catch(NcException& e)
    {
      e.what(); 
      return NC_ERR;
    }
  }

  delete [ ] matrix;
  gsl_rng_free(r);

  return 0;
}
