/*
  Small program to convert HADEX data from
       http://www.metoffice.gov.uk/hadobs/hadex

  into netCDF format.

  Author      : Graziano Giuliani
  Institution : ICTP
  Section     : ESP
  Date        : Wed Sep  4 16:44:31 CEST 2013
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netcdf.h>
#include <zlib.h>

#define DELTALAT 2.50
#define DELTALON 3.75
#define YSTART 1951
#define YEND 2003
#define VARNUM 28

typedef struct thevars {
  char name[16];
  char standard_name[64];
  char long_name[256];
  char units[32];
} myvar;

void getlinevals(char *line, float *vals, int nline, int nvals);
char *strndup(const char *s, size_t n);
int varindex(char *vname, myvar *vars);

float missing_value = -999.99;

int main(int argc, char *argv[])
{
  myvar variables[VARNUM] =
   {{"XXX","unknow","unknown","1"},
    {"CDD","consecutive_dry_days","Maximum number of consecutive days when precipitation < 1 mm","days"},
    {"CSDI","cold_spell_duration","Annual count when at least six consecutive days of min temperature < 10th percentile","days"},
    {"CWD","consecutive_wet_days","Maximum number of consecutive days when precipitation >= 1 mm","days"},
    {"DTR","diurnal_temperature_range","Mean difference between daily max and min temperature","C"},
    {"FD","frost_days","Annual count when daily minimum temperature < 0C","days"},
    {"GSL","growing_season_length","Annual (1st Jan to 31st Dec in NH, 1st July to 30th June in SH) count between first span of at least 6 days with TG > 5 C and first span after July 1 (January 1 in SH) of 6 days with TG < 5C (where TG is daily mean temperature)","days"},
    {"ID","ice_days","Annual count when daily maximum temperature < 0C","days"},
    {"PRCPTOT","annual_total_precipitation","Annual total precipitation from days >= 1mm","mm"},
    {"R10mm","heavy_precipitation_days","Annual count when precipitation >= 10 mm","days"},
    {"R20mm","very_heavy_precipitation_days","Annual count when precipitation >= 20 mm","days"},
    {"R95p","very_wet_days","Annual total precipitation from days > 95th percentile","mm"},
    {"R95pT","extreme_temperature_annual_contribution_from_very_wet_days","100 * R95p / PRCPTOT","%"},
    {"R99p","extremely_wet_days","Annual total precipitation from days > 99th percentile","mm"},
    {"RX1day","maximum_1_day_precip","Maximum 1 day precipitation","mm"},
    {"RX5day","maximum_5_days_precip","Maximum 5 day precipitation","mm"},
    {"SDII","simple_daily_intensity","The ratio of annual total precipitation to the number of wet days (>= 1 mm)","mm/day"},
    {"SU","summer_days","Annual count when daily max temperature > 25C","days"},
    {"TN10p","cold_night","Percentage of time when daily min temperature < 10th percentile","%"},
    {"TN90p","warm_night","Percentage of time when daily min temperature > 90th percentile","%"},
    {"TNn","mimimum_tmin","Minimum of minimum temperature","C"},
    {"TNx","maximum_tmin","Maximum of minimum temperature","C"},
    {"TR","tropical_nights","Annual count when daily min temperature >2 C","days"},
    {"TX10p","cold_days","Percentage of time when daily max temperature < 10th percentile","%"},
    {"TX90p","warm_days","Percentage of time when daily max temperature > 90th percentile","%"},
    {"TXn","minimum_tmax","Minimum of maximum temperature","C"},
    {"TXx","maximum_tmax","Maximum of maximum temperature","C"},
    {"WSDI","warm_spell_duration","Annual count when at least six consecutive days of max temperature > 90th percentile","days"} };

  double *lat, *lon, time;
  float *val;
  char filename[256], *pf;
  char vname[16];
  char *p, *q;
  char linein[1024];
  int nlat , nlon;
  int i, iy, iret;
  gzFile fp;
  int ncid, dimid[3], varid[4];
  size_t istart[3], icount[3];

  if ( argc < 2 )
  {
    fprintf(stderr, "No file to open\n");
    return 1;
  }
  fp = gzopen(argv[1], "r");
  if ( fp == Z_NULL )
  {
    fprintf(stderr, "Error opening file %s\n", argv[1]);
    return 2;
  }

  pf = strndup(argv[1],1024);
  p = strstr(pf,".");
  *p = 0;
  (void) snprintf(filename,1024,"%s%s",pf,".nc");
  iret = nc_create(filename,NC_CLOBBER|NC_NETCDF4|NC_CLASSIC_MODEL,&ncid);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error opening out file %s\n", filename);
    return 3;
  }

  p = strstr(pf,"_");
  p ++;
  q = strstr(p,"_");
  *q = 0;
  (void) snprintf(vname,16,"%s",p);

  nlat = (180.0/DELTALAT);
  nlon = (360.0/DELTALON);

  iret = nc_def_dim(ncid,"lon",nlon,dimid+2);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define lon dimension\n");
    return 4;
  }
  iret = nc_def_dim(ncid,"lat",nlat,dimid+1);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define lat dimension\n");
    return 4;
  }
  iret = nc_def_dim(ncid,"time",NC_UNLIMITED,dimid);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define time dimension\n");
    return 4;
  }
 
  iret = nc_def_var(ncid,"lat",NC_DOUBLE,1,dimid+1,varid);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define lat variable\n");
    return 5;
  }
  (void) nc_put_att_text(ncid,varid[0],"standard_name",
          strlen("latitude")+1,"latitude");
  (void) nc_put_att_text(ncid,varid[0],"units",
          strlen("degrees_north")+1,"degrees_north");
  iret = nc_def_var(ncid,"lon",NC_DOUBLE,1,dimid+2,varid+1);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define lon variable\n");
    return 5;
  }
  (void) nc_put_att_text(ncid,varid[1],"standard_name",
          strlen("longitude")+1,"longitude");
  (void) nc_put_att_text(ncid,varid[1],"units",
          strlen("degrees_east")+1,"degrees_east");
  iret = nc_def_var(ncid,"time",NC_DOUBLE,1,dimid,varid+2);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define time variable\n");
    return 5;
  }
  (void) nc_put_att_text(ncid,varid[2],"standard_name",
          strlen("time")+1,"time");
  (void) nc_put_att_text(ncid,varid[2],"units",
          strlen("years since 1951-06-15 00:00:00")+1,
          "years since 1951-06-15 00:00:00");
  iret = nc_def_var(ncid,vname,NC_FLOAT,3,dimid,varid+3);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error define variable\n");
    return 5;
  }
  iret = nc_def_var_deflate(ncid,*(varid+3), 1, 1, 9);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error setting compress level for variable\n");
    return 5;
  }

  i = varindex(vname,variables);
  (void) nc_put_att_text(ncid,varid[3],"standard_name",
          strlen(variables[i].standard_name)+1,variables[i].standard_name);
  (void) nc_put_att_text(ncid,varid[3],"long_name",
          strlen(variables[i].long_name)+1,variables[i].long_name);
  (void) nc_put_att_text(ncid,varid[3],"units",
          strlen(variables[i].units)+1,variables[i].units);
  (void) nc_put_att_float(ncid,varid[3],"_FillValue",NC_FLOAT,1,&missing_value);

  iret = nc_enddef(ncid);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error preparing file %s\n", filename);
    return 6;
  }

  lat = (double *) malloc(nlat*sizeof(double));
  lon = (double *) malloc(nlon*sizeof(double));
  val = (float *) malloc(nlon*nlat*sizeof(float));

  lat[0] = 90.0-DELTALAT/2.0;
  lon[0] = -180.0+DELTALON/2.0;
  for ( i = 1; i < nlat; i ++)
    lat[i] = lat[i-1] - DELTALAT;
  for ( i = 1; i < nlon; i ++)
    lon[i] = lon[i-1] + DELTALON;
  
  /* Skip first two lines */
  p = gzgets(fp,linein,1024);
  p = gzgets(fp,linein,1024);

  istart[0] = 0;
  icount[0] = nlat;
  iret = nc_put_vara_double(ncid,varid[0],istart,icount,lat);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error writing lat variable\n");
    return 7;
  }
  istart[0] = 0;
  icount[0] = nlon;
  iret = nc_put_vara_double(ncid,varid[1],istart,icount,lon);
  if ( iret != NC_NOERR )
  {
    fprintf(stderr, "Error writing lon variable\n");
    return 7;
  }

  for ( iy = YSTART; iy <= YEND; iy ++ )
  {
    for ( i = 0; i < nlat; i ++)
    {
      p = gzgets(fp,linein,1024);
      getlinevals(linein,val,i,nlon);
    }
    p = gzgets(fp,linein,1024);
    istart[0] = iy-YSTART;
    istart[1] = 0;
    istart[2] = 0;
    icount[0] = 1;
    icount[1] = nlat;
    icount[2] = nlon;
    time = iy-YSTART;
    iret = nc_put_vara_double(ncid,varid[2],istart,icount,&time);
    if ( iret != NC_NOERR )
    {
      fprintf(stderr, "Error writing time variable at timestep %d\n", iy);
      return 8;
    }
    iret = nc_put_vara_float(ncid,varid[3],istart,icount,val);
    if ( iret != NC_NOERR )
    {
      fprintf(stderr, "Error writing variable at timestep %d\n", iy);
      return 8;
    }
  }

  gzclose(fp);
  iret = nc_close(ncid);
  return 0;
}

void getlinevals(char *line, float *vals, int nline, int nvals)
{
  char *p, *n;
  int i;
  p = line;
  for ( i = 0; i < nvals; i++)
  {
    vals[nline*nvals+i] = strtod(p,&n);
    if ( vals[nline*nvals+i] < -900 ) vals[nline*nvals+i] = missing_value;
    p = n;
  }
}

int varindex(char *vname, myvar *vars)
{
  int i;
  for ( i = 1; i < VARNUM; i ++ )
  {
    if ( strncmp(vname,vars[i].name,16) == 0 ) return i;
  }
  return 0;
}
