/***************************************************************************
 *   Copyright (C) 2012 by Graziano Giuliani                               *
 *   ggiulian at ictp.it                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details. (see COPYING)            *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   LIC: GPL                                                              *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <netcdf.h>
#include <time.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#define SIZEOFDATE 64

int main(int argc, char *argv[])
{
  double x = 0.0, y = 0.0, dx, dy;
  double *matrix;
  int hx;
  const gsl_rng_type *T;
  char datetime[SIZEOFDATE];
  gsl_rng *r;
  int i , j, it;
  int nx , nt;
  int idims[3];
  size_t start[3], count[3];
  int errval , ncid, ivar;
  time_t xt;
  struct tm *tmp;

  if ( argc < 3 )
  {
    fprintf(stdout, "Not enough arguments.\n");
    fprintf(stdout, "Usage:\n");
    fprintf(stdout, "%s nx nt outfile\n", basename(argv[0]));
    fprintf(stdout, "where: \n");
    fprintf(stdout, "\tnx is the size of the problem\n");
    fprintf(stdout, "\tnt is the number of timesteps\n");
    fprintf(stdout, "\toutfile is the name of the output file\n");
    return 1;
  }

  nx = (int) strtol(argv[1], NULL, 0);
  nt = (int) strtol(argv[2], NULL, 0);
  if ( nx < 1 || nt < 1 )
  {
    fprintf(stderr, "NX must be greater than one\n");
    return -2;
  }

  /* Limit for the "walk" */
  hx = nx/2;

  fprintf(stdout, "NX = %d\n", nx);
  fprintf(stdout, "NT = %d\n", nt);

  /* Create the new file */
  errval = nc_create(argv[3], NC_CLOBBER, &ncid);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot write output file %s!\n", argv[3]);
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }

  /* GLOBAL attributes */

  errval = nc_put_att_text(ncid, NC_GLOBAL, "title", 28,
                           "Random walk of a log surface");
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Put att title!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }
  xt = time(NULL);
  tmp = localtime(&xt);
  if (tmp == NULL)
  {
    fprintf(stderr, "System error in getting time!\n");
    return -2;
  }
  (void) strftime(datetime, SIZEOFDATE, "Created on %c", tmp);
  errval = nc_put_att_text(ncid, NC_GLOBAL, "history", strlen(datetime)+1,
                           datetime);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Put att history!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }
  errval = nc_put_att_text(ncid, NC_GLOBAL, "purpose", 12, "Test GSL RNG");
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Put att purpose!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }

  /* Define the dimensions */

  errval = nc_def_dim(ncid, "X", nx, idims+2);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Define X Dimension!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }
  errval = nc_def_dim(ncid, "Y", nx, idims+1);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Define Y Dimension!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }
  errval = nc_def_dim(ncid, "time", NC_UNLIMITED, idims);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Define time Dimension!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }

  /* Now define variables ! */

  errval = nc_def_var(ncid, "randomwalk", NC_DOUBLE, 3, idims, &ivar);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Define randomwalk Variable!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }
  errval = nc_put_att_text(ncid, ivar, "units", 13, "dimensionless");
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Put att units!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }

  /* Put the file in WRITE DATA mode ending definitions */
  /* We can use nc_redef to get back to define mode     */

  errval = nc_enddef(ncid);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Exit Define Mode!\n");
    fprintf(stderr, "%s\n", nc_strerror(errval));
    return -1;
  }


  /* The DATA we are going to put in the file ! */

  matrix = (double *) malloc(nx*nx*sizeof(double));
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc(T);

  /* Loop on times */

  for (it = 0; it < nt; it++)
  {
    gsl_ran_dir_2d(r, &dx, &dy);
    x += dx;
    y += dy;
    for (i = 0; i < nx; i++)
    {
      for (j = 0; j < nx; j++)
      {
        matrix[j*nx+i] = log(pow((double)(i-hx)-x,2.0) + 
                             pow((double)(j-hx)-y,2.0)+0.1);
      }
    }

    /* Define the subset of the data in the netcdf file */

    start[0] = it;
    start[1] = 0;
    start[2] = 0;
    count[0] = 1;
    count[1] = nx;
    count[2] = nx;

    /* Write the slice into file */

    errval = nc_put_vara_double(ncid, ivar, start, count, matrix);
    if (errval != NC_NOERR)
    {
      fprintf(stderr, "Cannot Write var at step %d!\n", it);
      fprintf(stderr, "%s\n", nc_strerror(errval));
      return -1;
    }

    /* SYNC so as a reader can get the data as soon as
       they are available. This has a performance hit
       and shouldn't be left in production */

    errval = nc_sync(ncid);
    if (errval != NC_NOERR)
    {
      fprintf(stderr, "Error in Sync file!\n");
      fprintf(stderr, "%s\n", nc_strerror(errval));
      return -1;
    }

  }

  /* Finally, close the file */

  errval = nc_close(ncid);
  if (errval != NC_NOERR)
  {
    fprintf(stderr, "Cannot Close Output File!\n");
    return -1;
  }
  
  free(matrix);
  gsl_rng_free(r);

  return 0;
}
